# coding=utf-8
import tensorflow as tf
from tensorflow.python.eager import context

from tf_geometric.nn.kernel.map_reduce import aggregate_neighbors, sum_updater, sum_reducer, identity_updater
from tf_geometric.utils.graph_utils import add_self_loop_edge

CACHE_KEY_GCN_NORMED_EDGE = "gcn_normed_edge"


def gcn_norm_edge(edge_index, num_nodes, edge_weight=None, renorm=True, improved=False, cache: dict=None):

    if cache is not None:
        cached_data = cache.get(CACHE_KEY_GCN_NORMED_EDGE, None)
        if cached_data is not None:
            return cached_data

    # if cache is not None and CACHE_KEY_GCN_NORMED_EDGE in cache and cache[CACHE_KEY_GCN_NORMED_EDGE] is not None:
    #     return cache[CACHE_KEY_GCN_NORMED_EDGE]

    if edge_weight is None:
        edge_weight = tf.ones([tf.shape(edge_index)[1]], dtype=tf.float32)

    fill_weight = 2.0 if improved else 1.0

    if renorm:
        edge_index, edge_weight = add_self_loop_edge(edge_index, num_nodes, edge_weight=edge_weight, fill_weight=fill_weight)

    row, col = edge_index[0], edge_index[1]
    deg = tf.math.unsorted_segment_sum(edge_weight, row, num_segments=num_nodes)
    deg_inv_sqrt = tf.pow(deg, -0.5)
    deg_inv_sqrt = tf.where(
        tf.math.logical_or(tf.math.is_inf(deg_inv_sqrt), tf.math.is_nan(deg_inv_sqrt)),
        tf.zeros_like(deg_inv_sqrt),
        deg_inv_sqrt
    )

    normed_edge_weight = tf.gather(deg_inv_sqrt, row) * edge_weight * tf.gather(deg_inv_sqrt, col)

    if not renorm:
        edge_index, normed_edge_weight = add_self_loop_edge(edge_index, num_nodes, edge_weight=normed_edge_weight,
                                                            fill_weight=fill_weight)

    if cache is not None:
        cache[CACHE_KEY_GCN_NORMED_EDGE] = edge_index, normed_edge_weight

    return edge_index, normed_edge_weight


def gcn_cache_normed_edge(graph, renorm=True, improved=False, override=False):
    if override:
        graph.cache[CACHE_KEY_GCN_NORMED_EDGE] = None
    gcn_norm_edge(graph.edge_index, graph.num_nodes, graph.edge_weight, renorm, improved, graph.cache)


def gcn_mapper(repeated_x, neighbor_x, edge_weight=None):
    return neighbor_x * tf.expand_dims(edge_weight, 1)




def gcn(x, edge_index, edge_weight, kernel, bias=None, activation=None,
        renorm=True, improved=False, cache=None):
    """

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param kernel: Tensor, shape: [num_features, num_output_features], weight
    :param bias: Tensor, shape: [num_output_features], bias
    :param activation: Activation function to use.
    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        To use @tf_utils.function with gcn, you should cache the noremd edge information before the first call of the gcn.
        (1) If you're using OOP APIs tfg.layers.GCN:
            gcn_layer.cache_normed_edge(graph)
        (2) If you're using functional API tfg.nn.gcn:
            from tf_geometric.nn.conv.gcn import gcn_cache_normed_edge
            gcn_cache_normed_edge(graph)
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    num_nodes = tf.shape(x)[0]
    updated_edge_index, normed_edge_weight = gcn_norm_edge(edge_index, num_nodes, edge_weight, renorm, improved, cache)

    x = x @ kernel

    h = aggregate_neighbors(
        x, updated_edge_index, normed_edge_weight,
        gcn_mapper,
        sum_reducer,
        identity_updater,
        num_nodes=num_nodes
    )

    if bias is not None:
        h += bias

    if activation is not None:
        h = activation(h)

    return h






# coding=utf-8

import tensorflow as tf
from tf_geometric.nn.conv.chebynet import chebynet_norm_edge, chebynet
from tf_geometric.layers.kernel.map_reduce import MapReduceGNN


class ChebyNet(MapReduceGNN):
    """
    The chebyshev spectral graph convolutional operator from the
    `"Convolutional Neural Networks on Graphs with Fast Localized Spectral
    Filtering" <https://arxiv.org/abs/1606.09375>`_ paper

    """

    def __init__(self, units, K, lambda_max, activation=tf.nn.relu, use_bias=True, normalization_type='sym',
                 *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param K: Chebyshev filter size (default: '3").
        :param lambda_max:
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param activation: Activation function to use.
        :param normalization_type: The normalization scheme for the graph
            Laplacian (default: :obj:`"sym"`)

        """
        super().__init__(*args, **kwargs)
        self.units = units

        assert K > 0
        assert lambda_max is not None
        assert normalization_type in [None, 'sym', 'rw'], 'Invalid normalization'

        self.K = K

        self.use_bias = use_bias

        self.kernel = None
        self.bias = None

        self.activation = activation
        self.normalization_type = normalization_type
        self.lambda_max = lambda_max

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[self.K, num_features, self.units], initializer="glorot_uniform")
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units], initializer="zeros")

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return chebynet(x, edge_index, edge_weight, self.K, self.lambda_max, self.kernel, self.bias, self.activation,
                       self.normalization_type)



# coding=utf-8

from tf_geometric.nn.conv.gcn import gcn, gcn_cache_normed_edge
from tf_geometric.layers.kernel.map_reduce import MapReduceGNN
import tensorflow as tf

class GCN(MapReduceGNN):
    """
    Graph Convolutional Layer
    """

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[num_features, self.units],
                                      initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def __init__(self, units, activation=None, use_bias=True,
                 renorm=True, improved=False,
                 kernel_regularizer=None, bias_regularizer=None, *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
        :param improved: Whether use improved GCN or not.
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.units = units

        self.acvitation = activation
        self.use_bias = use_bias

        self.kernel = None
        self.bias = None

        self.renorm = renorm
        self.improved = improved

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

    def cache_normed_edge(self, graph, override=False):
        gcn_cache_normed_edge(graph, self.renorm, self.improved, override=override)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return gcn(x, edge_index, edge_weight, self.kernel, self.bias,
                   activation=self.acvitation, renorm=self.renorm, improved=self.improved, cache=cache)



import tensorflow as tf
from tensorflow import keras
from tf_geometric.nn.conv.graph_sage import mean_graph_sage, gcn_graph_sage, mean_pool_graph_sage, \
    max_pool_graph_sage, lstm_graph_sage
from tf_geometric.layers.kernel.map_reduce import MapReduceGNN


"""
The GraphSAGE operator from the `"Inductive Representation Learning on
Large Graphs" <https://arxiv.org/abs/1706.02216>`_ paper
"""
class MeanGraphSage(MapReduceGNN):

    def __init__(self, units, activation=tf.nn.relu, use_bias=True,
                 normalize=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.normalize = normalize

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        self.neighs_kernel = self.add_weight("neighs_kernel", shape=[num_features, self.units],
                                             initializer="glorot_uniform")
        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, self.units],
                                           initializer="glorot_uniform")

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units * 2], initializer="zeros")

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """
        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return mean_graph_sage(x, edge_index, edge_weight, self.neighs_kernel, self.self_kernel, self.bias,
                               self.activation, self.normalize)


class GCNGraphSage(MapReduceGNN):
    def __init__(self, units, activation=tf.nn.relu, use_bias=True,
                 normalize=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.normalize = normalize

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[num_features, self.units],
                                      initializer="glorot_uniform")

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units], initializer="zeros")

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return gcn_graph_sage(x, edge_index, edge_weight, self.kernel, self.bias, self.activation, self.normalize, cache=cache)


class MeanPoolGraphSage(MapReduceGNN):
    def __init__(self, units, activation=tf.nn.relu, use_bias=True,
                 normalize=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.normalize = normalize

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        self.mlp_kernel = self.add_weight("mlp_kernel", shape=[num_features, self.units*4],
                                          initializer="glorot_uniform")
        if self.use_bias:
            self.mlp_bias = self.add_weight("mlp_bias", shape=[self.units*4], initializer="zeros")
        # self.mlp_kernel = keras.layers.Dense(self.units, input_dim=2, use_bias=True, kernel_regularizer= tf.nn.l2_normalize, activation=tf.nn.relu)

        self.neighs_kernel = self.add_weight("neighs_kernel", shape=[self.units*4, self.units],
                                             initializer="glorot_uniform")
        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, self.units],
                                           initializer="glorot_uniform")

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units * 2], initializer="zeros")


    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return mean_pool_graph_sage(x, edge_index, edge_weight, self.mlp_kernel, self.neighs_kernel,
                                    self.self_kernel, self.mlp_bias, self.bias, self.activation,
                                    self.normalize)


class MaxPoolGraphSage(MapReduceGNN):
    def __init__(self, units, activation=tf.nn.relu, use_bias=True,
                 normalize=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.normalize = normalize

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        self.mlp_kernel = self.add_weight("mlp_kernel", shape=[num_features, self.units*4],
                                          initializer="glorot_uniform")
        if self.use_bias:
            self.mlp_bias = self.add_weight("mlp_bias", shape=[self.units*4], initializer="zeros")
        self.neighs_kernel = self.add_weight("neighs_kernel", shape=[self.units*4, self.units],
                                             initializer="glorot_uniform")
        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, self.units],
                                           initializer="glorot_uniform")

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units * 2], initializer="zeros")

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return max_pool_graph_sage(x, edge_index, edge_weight, self.mlp_kernel, self.neighs_kernel,
                                   self.self_kernel,
                                   self.mlp_bias, self.bias, self.activation, self.normalize)


class LSTMGraphSage(MapReduceGNN):

    def __init__(self, units, activation=tf.nn.relu, use_bias=True,
                 normalize=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.normalize = normalize

    def build(self, input_shape):
        x_shape = input_shape[0]
        num_features = x_shape[-1]

        self.lstm = tf.keras.layers.LSTM(self.units, kernel_regularizer=tf.nn.l2_normalize, return_sequences=True)
        self.neighs_kernel = self.add_weight("neighs_kernel", shape=[self.units, self.units],
                                             initializer="glorot_uniform")
        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, self.units],
                                           initializer="glorot_uniform")

        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units * 2], initializer="zeros")


    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return lstm_graph_sage(x, edge_index, edge_weight, self.lstm, self.neighs_kernel,
                                       self.self_kernel, self.bias, self.activation,
                                      self.normalize)




# coding=utf-8

import tensorflow as tf
from tf_geometric.nn.conv.tagcn import tagcn
from tf_geometric.layers.kernel.map_reduce import MapReduceGNN


class TAGCN(MapReduceGNN):
    """
    The topology adaptive graph convolutional networks operator from the
     `"Topology Adaptive Graph Convolutional Networks"
     <https://arxiv.org/abs/1710.10370>`_ paper
    """

    def __init__(self, units, K=3, activation=tf.nn.relu, use_bias=True,
                 renorm=False, improved=False, *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param K: Number of hops (default: '3").
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
        :param improved: Whether use improved GCN or not.

        """
        super().__init__(*args, **kwargs)
        self.units = units
        assert K > 0

        self.K = K

        self.activation = activation
        self.use_bias = use_bias

        self.kernel = None
        self.bias = None

        self.renorm = renorm
        self.improved = improved

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[num_features * (self.K + 1), self.units],
                                      initializer="glorot_uniform")
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units], initializer="zeros")

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return tagcn(x, edge_index, edge_weight, self.K, self.kernel, self.bias, self.activation, self.renorm,
                     self.improved, cache)



# coding=utf-8

import numpy as np
import os

import networkx as nx
from tf_geometric.data.graph import Graph
from tf_geometric.data.dataset import DownloadableDataset
import json

from tf_geometric.utils.data_utils import load_cache
from tf_geometric.utils.graph_utils import convert_edge_to_directed


class PPIDataset(DownloadableDataset):

    def __init__(self, dataset_root_path=None):
        super().__init__(dataset_name="PPI",
                         download_urls=[
                             "https://data.dgl.ai/dataset/ppi.zip",
                             "https://github.com/CrawlScript/gnn_datasets/raw/master/PPI/ppi.zip"
                         ],
                         download_file_name="ppi.zip",
                         cache_name="cache.p",
                         dataset_root_path=dataset_root_path,
                         )

    def process(self):

        splits = ["train", "valid", "test"]

        split_data_dict = {
            split: [] for split in splits
        }

        for split in split_data_dict.keys():
            split_graph_ids = np.load(os.path.join(self.raw_root_path, "{}_graph_id.npy".format(split)))
            split_features = np.load(os.path.join(self.raw_root_path, "{}_feats.npy".format(split))).astype(np.float32)
            split_labels = np.load(os.path.join(self.raw_root_path, "{}_labels.npy".format(split))).astype(np.int32)

            nx_graph_path = os.path.join(self.raw_root_path, "{}_graph.json".format(split))
            with open(nx_graph_path, "r", encoding="utf-8") as f:
                nx_graph = nx.DiGraph(nx.json_graph.node_link_graph(json.load(f)))

            split_unique_graph_ids = sorted(set(split_graph_ids))

            for graph_id in split_unique_graph_ids:
                mask_indices = np.where(split_graph_ids == graph_id)[0]

                min_node_index = np.min(mask_indices)

                edge_index = nx_graph.subgraph(mask_indices).edges
                edge_index = np.array(edge_index).T - min_node_index

                edge_index, _ = convert_edge_to_directed(edge_index)

                graph = Graph(
                    x=split_features[mask_indices],
                    edge_index=edge_index,
                    y=split_labels[mask_indices]
                )
                split_data_dict[split].append(graph)
                # print("split: ", split)

        processed_data = [split_data_dict[split] for split in splits]
        return processed_data



# coding=utf-8

import tensorflow as tf
import warnings


def tf_func_warn(func):
    warnings.warn("Using @tf_utils.function to speed-up functions is only available for TensorFlow 2.x. Upgrade your TensorFlow to 2.x and the performance can be improved dramatically with @tf_utils.function")
    return func


# disable tf.function for tf 1
if tf.__version__[0] == "1":
    function = tf_func_warn
else:
    function = tf.function





# coding=utf-8
import tensorflow as tf

from tf_geometric.nn.kernel.map_reduce import aggregate_neighbors, sum_updater, sum_reducer, identity_updater
from tf_geometric.nn.kernel.segment import segment_softmax
from tf_geometric.nn.conv.gcn import gcn_mapper
from tf_geometric.utils.graph_utils import add_self_loop_edge


# follow Transformer-Style Attention
# Attention is all you need
def gat(x, edge_index,
        query_kernel, query_bias, query_activation,
        key_kernel, key_bias, key_activation,
        kernel, bias=None, activation=None, num_heads=1, drop_rate=0.0, training=False):
    """

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param query_kernel: Tensor, shape: [num_features, num_query_features], weight for Q in attention
    :param query_bias: Tensor, shape: [num_query_features], bias for Q in attention
    :param query_activation: Activation function for Q in attention.
    :param key_kernel: Tensor, shape: [num_features, num_key_features], weight for K in attention
    :param key_bias: Tensor, shape: [num_key_features], bias for K in attention
    :param key_activation: Activation function for K in attention.
    :param kernel: Tensor, shape: [num_features, num_output_features], weight
    :param bias: Tensor, shape: [num_output_features], bias
    :param activation: Activation function to use.
    :param num_heads: Number of attention heads.
    :param drop_rate: Dropout rate.
    :param training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """


    num_nodes = tf.shape(x)[0]

    # self-attention
    edge_index, edge_weight = add_self_loop_edge(edge_index, num_nodes)

    row, col = edge_index[0], edge_index[1]

    Q = query_activation(x @ query_kernel + query_bias)
    Q = tf.gather(Q, row)

    K = key_activation(x @ key_kernel + key_bias)
    K = tf.gather(K, col)

    V = x @ kernel

    # xxxxx_ denotes the multi-head style stuff
    Q_ = tf.concat(tf.split(Q, num_heads, axis=-1), axis=0)
    K_ = tf.concat(tf.split(K, num_heads, axis=-1), axis=0)
    V_ = tf.concat(tf.split(V, num_heads, axis=-1), axis=0)
    edge_index_ = tf.concat([edge_index + i * num_nodes for i in range(num_heads)], axis=1)

    att_score_ = tf.reduce_sum(Q_ * K_, axis=-1)
    normed_att_score_ = segment_softmax(att_score_, edge_index_[0], num_nodes * num_heads)

    if training and drop_rate > 0.0:
        normed_att_score_ = tf.compat.v2.nn.dropout(normed_att_score_, drop_rate)

    h_ = aggregate_neighbors(
        V_, edge_index_, normed_att_score_,
        gcn_mapper,
        sum_reducer,
        identity_updater
    )

    h = tf.concat(tf.split(h_, num_heads, axis=0), axis=-1)

    if bias is not None:
        h += bias

    if activation is not None:
        h = activation(h)

    return h










# coding=utf-8

import tensorflow as tf
from tf_geometric.nn.kernel.segment import segment_op_with_pad


def identity_mapper(repeated_x, neighbor_x, edge_weight=None):
    return neighbor_x


def neighbor_count_mapper(repeated_x, neighbor_x, edge_weight=None):
    return tf.ones([neighbor_x.shape[0], 1])


def sum_reducer(neighbor_msg, node_index, num_nodes=None):
    return tf.math.unsorted_segment_sum(neighbor_msg, node_index, num_segments=num_nodes)


def sum_updater(x, reduced_neighbor_msg):
    return x + reduced_neighbor_msg


def identity_updater(x, reduced_neighbor_msg):
    return reduced_neighbor_msg


def mean_reducer(neighbor_msg, node_index, num_nodes=None):
    return tf.math.unsorted_segment_mean(neighbor_msg, node_index, num_segments=num_nodes)


def max_reducer(neighbor_msg, node_index, num_nodes=None):
    if num_nodes is None:
        num_nodes = tf.reduce_max(node_index) + 1
    # return tf.math.unsorted_segment_max(neighbor_msg, node_index, num_segments=num_nodes)
    # max_x = tf.math.unsorted_segment_max(x, node_graph_index, num_segments=num_graphs)
    max_neighbor_msg = segment_op_with_pad(tf.math.segment_max, neighbor_msg, node_index, num_segments=num_nodes)
    return max_neighbor_msg


def aggregate_neighbors(x, edge_index, edge_weight=None, mapper=identity_mapper,
                        reducer=sum_reducer, updater=sum_updater, num_nodes=None):
    """
    :param x:
    :param edge_index:
    :param mapper: (features_of_node, features_of_neighbor_node, edge_weight) => neighbor_msg
    :param reducer: (neighbor_msg, node_index) => reduced_neighbor_msg
    :param updater: (features_of_node, reduced_neighbor_msg, num_nodes) => aggregated_node_features
    :param num_nodes: Number of nodes.
    :return:
    """

    if tf.shape(edge_index)[0] == 0:
        return x

    row, col = edge_index[0], edge_index[1]

    repeated_x = tf.gather(x, row)
    neighbor_x = tf.gather(x, col)

    neighbor_msg = mapper(repeated_x, neighbor_x, edge_weight=edge_weight)

    if num_nodes is None:
        num_nodes = tf.shape(x)[0]

    reduced_msg = reducer(neighbor_msg, row, num_nodes=num_nodes)
    udpated = updater(x, reduced_msg)

    return udpated









# coding=utf-8
import tensorflow as tf

from tf_geometric.nn.kernel.segment import segment_count, segment_op_with_pad


def mean_pool(x, node_graph_index, num_graphs=None):
    if num_graphs is None:
        num_graphs = tf.reduce_max(node_graph_index) + 1
    num_nodes_of_graphs = segment_count(node_graph_index, num_segments=num_graphs)
    sum_x = tf.math.unsorted_segment_sum(x, node_graph_index, num_segments=num_graphs)
    return sum_x / (tf.cast(tf.expand_dims(num_nodes_of_graphs, -1), tf.float32) + 1e-8)


def sum_pool(x, node_graph_index, num_graphs=None):
    if num_graphs is None:
        num_graphs = tf.reduce_max(node_graph_index) + 1
    sum_x = tf.math.unsorted_segment_sum(x, node_graph_index, num_segments=num_graphs)
    return sum_x


def max_pool(x, node_graph_index, num_graphs=None):
    if num_graphs is None:
        num_graphs = tf.reduce_max(node_graph_index) + 1
    # max_x = tf.math.unsorted_segment_max(x, node_graph_index, num_segments=num_graphs)
    max_x = segment_op_with_pad(tf.math.segment_max, x, node_graph_index, num_segments=num_graphs)
    return max_x


def min_pool(x, node_graph_index, num_graphs=None):
    if num_graphs is None:
        num_graphs = tf.reduce_max(node_graph_index) + 1
    # min_x = tf.math.unsorted_segment_min(x, node_graph_index, num_segments=num_graphs)
    min_x = segment_op_with_pad(tf.math.segment_min, x, node_graph_index, num_segments=num_graphs)
    return min_x










# coding=utf-8
import tensorflow as tf

from tf_geometric.nn.conv.gat import gat
from tf_geometric.layers.kernel.map_reduce import MapReduceGNN


class GAT(MapReduceGNN):

    def __init__(self, units,
                 attention_units=None,
                 activation=None,
                 use_bias=True,
                 num_heads=1,
                 query_activation=tf.nn.relu,
                 key_activation=tf.nn.relu,
                 drop_rate=0.0,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param attention_units: Positive integer, dimensionality of the output space for Q and K in attention.
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param num_heads: Number of attention heads.
        :param query_activation: Activation function for Q in attention.
        :param key_activation: Activation function for K in attention.
        :param drop_rate: Dropout rate.
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.units = units
        self.attention_units = units if attention_units is None else attention_units
        self.drop_rate = drop_rate

        self.query_kernel = None
        self.query_bias = None
        self.query_activation = query_activation

        self.key_kernel = None
        self.key_bias = None
        self.key_activation = key_activation

        self.kernel = None
        self.bias = None

        self.acvitation = activation
        self.use_bias = use_bias
        self.num_heads = num_heads

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        self.query_kernel = self.add_weight("query_kernel", shape=[num_features, self.attention_units],
                                            initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        self.query_bias = self.add_weight("query_bias", shape=[self.attention_units],
                                          initializer="zeros", regularizer=self.bias_regularizer)

        self.key_kernel = self.add_weight("key_kernel", shape=[num_features, self.attention_units],
                                          initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        self.key_bias = self.add_weight("key_bias", shape=[self.attention_units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

        self.kernel = self.add_weight("kernel", shape=[num_features, self.units],
                                      initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def call(self, inputs, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index]
        :return: Updated node features (x), shape: [num_nodes, units]
        """
        x, edge_index = inputs

        return gat(x, edge_index,
                   self.query_kernel, self.query_bias, self.query_activation,
                   self.key_kernel, self.key_bias, self.key_activation,
                   self.kernel, self.bias, self.acvitation,
                   num_heads=self.num_heads,
                   drop_rate=self.drop_rate,
                   training=training)


# coding=utf-8

from tf_geometric.nn.conv.sgc import sgc
from tf_geometric.layers.kernel.map_reduce import MapReduceGNN

class SGC(MapReduceGNN):
    """
    The simple graph convolutional operator from the `"Simplifying Graph
    Convolutional Networks" <https://arxiv.org/abs/1902.07153>`_ paper
    """

    def __init__(self, units, k=1, use_bias = True, renorm=True, improved=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        """
        :param units: Size of each output sample..
        :param k: Number of hops.(default: :obj:`1`)
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
        :param improved: Whether use improved GCN or not.
        """

        self.units = units
        self.use_bias = use_bias
        self.renorm = renorm
        self.improved = improved
        self.K = k
        self.kernel = []
        self.bias = []

    def build(self, input_shape):

        x_shape = input_shape[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[num_features, self.units], initializer="glorot_uniform")
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units], initializer="zeros")

    def call(self, inputs, cache=None, training=None, mask=None):
        """
        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, num_units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return sgc(x, edge_index, edge_weight, self.K, self.kernel, self.bias, self.renorm, self.improved, cache)





# coding=utf-8
import tensorflow as tf
from tensorflow import keras
from tf_geometric.nn.pool.common_pool import mean_pool


class MeanPool(keras.Model):

    def call(self, inputs, training=None, mask=None):
        if len(inputs) == 2:
            x, node_graph_index = inputs
            num_graphs = None
        else:
            x, node_graph_index, num_graphs = inputs

        return mean_pool(x, node_graph_index, num_graphs)


# coding=utf-8

# nn package contains functional APIs for tf_geometric
from tf_geometric.nn.conv.gcn import *
from tf_geometric.nn.conv.gat import *
from tf_geometric.nn.kernel.map_reduce import *
from tf_geometric.nn.pool.common_pool import *
from tf_geometric.nn.pool.topk_pool import *
from tf_geometric.nn.conv.chebynet import *
from tf_geometric.nn.conv.sgc import *
from tf_geometric.nn.conv.tagcn import *
from tf_geometric.nn.conv.graph_sage import *





from tf_geometric.layers.conv.gcn import GCN
from tf_geometric.layers.conv.gat import GAT
from tf_geometric.layers.kernel.map_reduce import MapReduceGNN
from tf_geometric.layers.pool.mean_pool import MeanPool
from tf_geometric.layers.conv.gin import GIN
from tf_geometric.layers.conv.graph_sage import MeanGraphSage, MeanPoolGraphSage, MaxPoolGraphSage, GCNGraphSage, LSTMGraphSage
from tf_geometric.layers.conv.sgc import SGC
from tf_geometric.layers.conv.tagcn import TAGCN
from tf_geometric.layers.conv.chebynet import ChebyNet


import tensorflow as tf
from tf_geometric.nn.conv.gcn import gcn_mapper
from tf_geometric.utils.graph_utils import remove_self_loop_edge, add_self_loop_edge, get_laplacian
from tf_geometric.nn.kernel.map_reduce import aggregate_neighbors, sum_reducer, identity_updater


def chebynet_norm_edge(edge_index, num_nodes, edge_weight, lambda_max, normalization_type):
    edge_index, edge_weight = remove_self_loop_edge(edge_index, edge_weight)

    updated_edge_index, updated_edge_weight = get_laplacian(edge_index, edge_weight, normalization_type,
                                                            num_nodes)

    scaled_edge_weight = (2.0 * updated_edge_weight) / lambda_max

    assert edge_weight is not None

    return updated_edge_index, scaled_edge_weight


def chebynet(x, edge_index, edge_weight, K, lambda_max, kernel, bias=None, activation=None, normalization_type=None):
    num_nodes = x.shape[0]
    norm_edge_index, norm_edge_weight = chebynet_norm_edge(edge_index, num_nodes, edge_weight, lambda_max,
                                                           normalization_type=normalization_type)

    T0_x = x
    T1_x = x
    out = tf.matmul(T0_x, kernel[0])

    if K > 1:
        T1_x = aggregate_neighbors(x, norm_edge_index, norm_edge_weight, gcn_mapper, sum_reducer, identity_updater)
        out += tf.matmul(T1_x, kernel[1])

    for i in range(2, K):
        T2_x = aggregate_neighbors(T1_x, norm_edge_index, norm_edge_weight, gcn_mapper, sum_reducer,
                                   identity_updater)  ##L^T_{k-1}(L^)
        T2_x = 2.0 * T2_x - T0_x
        out += tf.matmul(T2_x, kernel[i])

        T0_x, T1_x = T1_x, T2_x

    if bias is not None:
        out += bias

    if activation is not None:
        out += activation(out)

    return out



# coding=utf-8
import os
# multi-gpu ids
os.environ["CUDA_VISIBLE_DEVICES"] = "4,5"
import tf_geometric as tfg
from tf_geometric.layers import GCN
from tensorflow.keras.regularizers import L1L2
import tensorflow as tf


graph, (train_index, valid_index, test_index) = tfg.datasets.CoraDataset().load_data()
num_classes = graph.y.max() + 1
learning_rate = 1e-2
l2_coef = 5e-4



# custom network
class GCNNetwork(tf.keras.Model):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gcn0 = GCN(16, activation=tf.nn.relu, kernel_regularizer=L1L2(l2=l2_coef))
        self.gcn1 = GCN(num_classes, kernel_regularizer=L1L2(l2=l2_coef))
        self.dropout = tf.keras.layers.Dropout(0.5)

    def call(self, inputs, training=None, mask=None):
        x, edge_index = inputs
        print(x.shape, edge_index.shape)
        x = x[0]
        edge_index = edge_index[0]

        h = self.dropout(x, training=training)
        h = self.gcn0([h, edge_index], training=training)
        h = self.dropout(h, training=training)
        h = self.gcn1([h, edge_index], training=training)
        return h

# prepare a generator and a dataset for distributed training
def create_batch_generator():
    while True:
        # return a tupe (X, y) as the data for a batch
        # yield (graph.x, graph.edge_index), graph.y
        import numpy as np
        # yield (graph.x, graph.edge_index[:, :np.random.randint(0, 10)]), graph.y
        # yield ([graph.x], [graph.edge_index[:, :np.random.randint(0, 10)]]), [graph.y]
        # yield ([graph.x] * 2, [graph.edge_index[:, :np.random.randint(0, 10)]] * 2), [graph.y] * 2
        # yield ([graph.x] * 8, [graph.edge_index[:, :np.random.randint(0, 10)] for _ in range(8)]), [graph.y] * 8
        x = tf.convert_to_tensor([graph.x] * 8)
        edge_index = tf.ragged.constant(
            [graph.edge_index[:, :np.random.randint(0, 10)] for _ in range(8)],
            dtype=tf.int32
        )
        y = tf.convert_to_tensor([graph.y] * 8)
        return (x, edge_index), y

dataset = tf.data.Dataset.from_generator(
    create_batch_generator,
    output_types=((tf.float32, tf.int32), tf.int32),
    # shape for (graph.x, graph.edge_index), graph.y
    # output_shapes=((tf.TensorShape([None, graph.x.shape[1]]), tf.TensorShape([2, None])), tf.TensorShape([None]))
    output_shapes=((tf.TensorShape([None, None, graph.x.shape[1]]), tf.TensorShape([None, 2, None])), tf.TensorShape([None, None]))
)

strategy = tf.distribute.MirroredStrategy()
dataset = strategy.experimental_distribute_datasets_from_function(

)


# create a distributed model, which uses all GPUs defined by CUDA_VISIBLE_DEVICES
with tf.distribute.MirroredStrategy().scope():
    model = GCNNetwork()

# custom loss function
def masked_cross_entropy(y_true, logits):

    y_true = y_true[0]
    y_true = tf.cast(y_true, tf.int32)
    masked_logits = tf.gather(logits, train_index)
    masked_labels = tf.gather(y_true, train_index)
    losses = tf.nn.softmax_cross_entropy_with_logits(
        logits=masked_logits,
        labels=tf.one_hot(masked_labels, depth=num_classes)
    )
    # print("loss ---- ", tf.reduce_mean(losses))

    return tf.reduce_mean(losses)


model.compile(
    optimizer=tf.keras.optimizers.Adam(learning_rate=1e-2),
    loss=masked_cross_entropy
)


def evaluate():
    import numpy as np
    logits = model([np.array([graph.x]), np.array([graph.edge_index])])
    masked_logits = tf.gather(logits, test_index)
    masked_labels = tf.gather(graph.y, test_index)

    y_pred = tf.argmax(masked_logits, axis=-1, output_type=tf.int32)
    corrects = tf.cast(tf.equal(masked_labels, y_pred), tf.float32)
    accuracy = tf.reduce_mean(corrects)
    return accuracy.numpy()

    # accuracy_m = tf.keras.metrics.Accuracy()
    # accuracy_m.update_state(masked_labels, y_pred)
    # return accuracy_m.result().numpy()
    #

# custom a callback function for evaluation during training
class EvaluationCallback(tf.keras.callbacks.Callback):
    def on_epoch_end(self, epoch, logs=None):
        # print("====")
        return
        if epoch % 100 == 0:
            test_accuracy = evaluate()
            print("epoch = {}\ttest_accuracy = {}".format(epoch, test_accuracy))


model.fit(dataset, steps_per_epoch=1, epochs=10000, callbacks=[EvaluationCallback()], verbose=True)


test_accuracy = evaluate()
print("test_accuracy = {}".format(test_accuracy))